const fs = require('fs')
const os = require('os')
const md5 = require('md5')
const { promisify } = require('util')
const PythonWorker = require('./pythonWorker')

const readdir = promisify(fs.readdir)
const stat = promisify(fs.stat)

/**
 * INTERFACE BACKEND
 *
 * Communication with workers (for now only Python)
 * The workers are implemented as REPL clients that
 * communicate with JSON.
 *
 * Names:
 *  * Worker: a Python file that implements a REPL client (e.g. serial)
 *  * Port:
 *  * Connection:
 */

const WORKER_DIR = `${__dirname}/python_workers`
const HOST = os.hostname()

const state = {
  interfaces: [],
  ports: [],
  connections: [],
  lastScan: {
    workers: Date.now() - 100000,
    ports: Date.now() - 100000,
    connections: Date.now() - 100000
  }
}

const typeDefs = `
  type Interface {
    id: ID!
    interfaceName: String!
    workerScript: WorkerScript!
    workerProcess: WorkerProcess
    ports: [Port]!
    connections: [Connection]!
    options: [Option!]
  }

  type WorkerScript {
    path: String!
    mtime: DateTime!
    updated: DateTime
  }

  type WorkerProcess {
    pid: Int!
    killed: Boolean!
    signalCode: String
    exitCode: Int
    spawnfile: String!
    spawnargs: [String]!
  }

  type Option {
    name: String!
    type: String!
    description: String
    values: [String!]
  }

  type Port {
    id: String!
    interfaceName: String!
    host: String!
    device: String
    name: String
    description: String
  }

  type Connection {
    id: ID!
    port: Port!
    workerProcess: WorkerProcess!
  }

  input ConnectionCommand {
    type: String!
    data: String
    options: String
  }

  extend type Query {
    interfaces: [Interface]!
    interface(id: ID!): Interface!
    ports: [Port]!
    port(id: ID!): Port!
    connections: [Connection]!
    connection(id: ID!): Connection!
  }

  extend type Mutation {
    connect(portId: ID!): Connection!
    endConnection(connectionId: ID!): Connection!
    killConnection(connectionId: ID!): Connection!
    sendCommand(connectionId: ID!, command: ConnectionCommand!): String!
  }
`

/**
 * INTERFACE SECTION
 */
async function findInterfaces () {
  // 1. Don't check more frequently than once per second.
  // if (state.lastScan.workers + 1000 > Date.now()) return null
  // state.lastScan.workers = Date.now()

  // 2. Find all files in ./python_workers that end in _worker.py
  const fileNames = await readdir(WORKER_DIR)
  const workerFiles = fileNames.filter(fileName =>
    fileName.includes('_worker.py')
  )

  // 3. For every worker script
  const workerPromises = workerFiles.map(async workerFile => {
    const interfaceName = workerFile.replace(/_worker\.py/, '')
    const path = `${WORKER_DIR}/${workerFile}`
    // a. Find out if it was modified.
    const { mtime } = await stat(path)
    const workerScript = { path, mtime, updated: null }
    const foundInterface = state.interfaces.find(
      iface => iface.interfaceName === interfaceName
    )
    if (foundInterface) {
      // b. If it was modified, save the modification time.
      if (foundInterface.workerScript.mtime < mtime) {
        foundInterface.workerScript.updated = mtime
      }
      return
    }
    // c. Spawn a new worker connection.
    const workerProcess = new PythonWorker(workerScript)
    const { data, error } = await workerProcess.spawn()
    if (error) throw new Error(error)
    // d. Save the worker in the state.
    state.interfaces.push({
      id: md5(`${interfaceName}${workerScript}${mtime}`),
      interfaceName,
      workerScript,
      workerProcess,
      options: await options(workerProcess)
    })
  })
  await Promise.all(workerPromises)
}

async function interfaces (parent, args, context, info) {
  await findInterfaces()
  return state.interfaces
}

async function iface (parent, { id }, context, info) {
  await findInterfaces()
  const iface = state.interfaces.find(
    iface => iface.id === id || iface.interfaceName === interfaceName
  )
  if (!iface) {
    throw new Error(`Worker id=${id}, interfaceName=${interfaceName} not found`)
  }
  return {
    ...iface,
    workerProcess
  }
}

async function options (workerProcess) {
  const { data, error } = await workerProcess.send({
    type: 'options'
  })
  if (error) throw new Error(error)
  return data
}

function workerProcess (parent, args, context, info) {
  const {
    killed,
    exitCode,
    signalCode,
    spawnargs,
    spawnfile,
    pid
  } = parent.workerProcess.pythonShell
  return {
    pid,
    killed,
    exitCode,
    signalCode,
    spawnfile,
    spawnargs
  }
}

/**
 * PORTS SECTION
 */
async function findPorts () {
  // 1. Make sure, workers are updated.
  await findInterfaces()

  // 2. Don't check more frequently than once per second.
  // if (state.lastScan.ports + 1000 > Date.now()) return null
  // state.lastScan.ports = Date.now()

  // 3. Loop through all workers to find available ports.
  const portsPromises = state.interfaces.map(async iface => {
    // a) Ask interface for ports.
    const { data, error } = await iface.workerProcess.send({
      type: 'ports'
    })
    if (error) throw new Error(error)
    // b) Add all ports that are not in the list.
    data.forEach(port => {
      const id = port.name || port.device
      if (state.ports.find(port => port.id === id)) return null
      const newPort = {
        id,
        interface: iface,
        interfaceName: iface.interfaceName,
        host: HOST,
        ...port
      }
      state.ports.push(newPort)
    })
  })
  await Promise.all(portsPromises)
}

async function ports (parent, args, context, info) {
  await findPorts()

  if (parent) {
    return state.ports.filter(
      port => port.interfaceName === parent.interfaceName
    )
  } else {
    return state.ports
  }
}

async function port (parent, { id }, context, info) {
  await findPorts()
  const port = state.ports.find(port => port.id === id)
  return port
}

/**
 * CONNECTION SECTION
 */
async function connect (parent, { portId }, ctx, info) {
  await findPorts()
  const port = state.ports.find(port => port.id === portId)
  if (!port) throw new Error(`Port ${portId} not found`)

  const id = md5(port.interfaceName + port.device)
  if (state.connections.find(connection => connection.id === id)) {
    throw new Error('already connected.')
  }
  const workerProcess = new PythonWorker(port.interface.workerScript)
  const spawnData = await workerProcess.spawn()
  if (spawnData.error) throw new Error(spawnData.error)
  const connection = {
    id,
    port,
    workerProcess
  }
  const connectionData = await workerProcess.send({
    type: 'connect',
    device: port.device
  })
  if (connectionData.error) throw new Error(connectionData.error)
  state.connections.push(connection)
  return connection
}

async function connections (parent, args, ctx, info) {
  if (parent) {
    return state.connections.filter(
      connection => connection.port.interfaceName === parent.interfaceName
    )
  } else {
    return state.connections
  }
}

async function connection (parent, { id }, context, info) {
  const connection = state.connections.find(connection => connection.id === id)
  return connection
}

async function sendCommand (parent, { connectionId, command }, ctx, info) {
  const connection = state.connections.find(
    connection => connection.id === connectionId
  )
  if (!connection) throw new Error('Connection not found.')
  const { data, error } = await connection.workerProcess.send({ ...command })
  if (error) throw new Error(JSON.stringify(error))
  if (!data.response) {
    return JSON.stringify(data)
  } else {
    return data.response
  }
}

async function endConnection (parent, { connectionId }, ctx, info) {
  const connectionIndex = state.connections.findIndex(
    connection => connection.id === connectionId
  )
  if (connectionIndex < 0) throw new Error('Connection not found.')
  const connection = state.connections[connectionIndex]

  const { data, error } = await connection.workerProcess.end()
  if (error) throw new Error(JSON.stringify(error))
  if (data) throw new Error(`Process ended with exit code ${data}`)
  state.connections.splice(connectionIndex, 1)
  return connection
}

async function killConnection (parent, { connectionId, signal = 9 }, ctx, info) {
  const connectionIndex = state.connections.findIndex(
    connection => connection.id === connectionId
  )
  if (connectionIndex < 0) throw new Error('Connection not found.')
  const connection = state.connections[connectionIndex]

  const { data, error } = await connection.workerProcess.kill(signal)
  console.log({ data, error })
  if (error) throw new Error(JSON.stringify(error))
  if (data) throw new Error(`Received code ${data}`)
  state.connections.splice(connectionIndex, 1)
  return connection
}

const resolvers = {
  Query: {
    interfaces,
    interface: iface,
    ports,
    port,
    connections,
    connection
  },
  Mutation: {
    connect,
    sendCommand,
    endConnection,
    killConnection
  },
  Interface: {
    ports,
    connections,
    workerProcess
  },
  Connection: {
    workerProcess
  }
}

module.exports = { typeDefs, resolvers }
